/*******************************************************************
********************************************************************
********************************************************************
*****                                                          *****
*****        L A B C E N T E R    E L E C T R O N I C S        *****
*****                                                          *****
*****       LABCENTER INTEGRATED SIMULATION ARCHITECTURE       *****
*****                                                          *****
*****          PIC1687* Driver for LCD in 4-bit mode           *****
*****                                                          *****
********************************************************************
********************************************************************/
// These driver routines implement the functionality necessary to drive
// a HD44780 controlled Alphameric LCD in 4-bit mode.
// PORT C pins 4-7 are used for data transmission and PORT B pins 0-2 are used 
// for command control.

// The Custom Graphic characters are written in 8-bytes blocks to the CGRAM. Subsequent
// to writing they are accessed via an index in DDRAM - that is, the custom graphic located
// at CGRAM address 40h-47h is accessed by writing to character font location 0x00 in DDRAM,
// the custom graphic located at 48-4F is accessed by writing to character font location 0x01
// in DDRAM and so on.

#include "lcd4bit.h"
#include "p18f6585.h"

void lcd_init ()
// Initialise the LCD Display. 
 { 
   Delay10TCYx(110);
   TRISD = PORTD = 0xFF;					  // inicializirane kato vhodove
   wrcmd(LCD_SETFUNCTION);               	// 4-bit mode - 2 line - 5x7 font. 
   wrcmd(LCD_SETVISIBLE+0x04);                // Display on, no cursor - no blink.
   wrcmd(LCD_SETMODE+0x03);                   // Automatic Increment - No Display shift.

   //wrcmd(LCD_SETMODE+0x03);                   // Automatic Increment - Display shift.
   //wrcmd(LCD_SHIFT+0x08);						// shift left
			  
   wrcmd(LCD_SETDDADDR+0x0F);                      // Address DDRAM with 0 offset 80h.
   }
 
void clearscreen ()
// Clear the LCD Screen and reset
// initial position.
 { 
	wrcmd(LCD_CLS);
    wrcmd(LCD_SETDDADDR+0x0F);
 }

void wrcmd (char cmdcode)
// Write a command to the LCD display.
// In 4-bit mode we send the MSN first and then
// the LSN. We then call the wait routine to hold
// until the busy flag is cleared.
 { TRISD = 0;
   PORTD = (cmdcode & 0xF0);   				  // Get the most significant nibble first.
   PORTE = TRISE = 0; 						  // inic. kato izhod i nulirane
   PORTE  = LCD_CMD_WR;                       // Specify a command write operation.
   PORTE |= E_PIN_MASK;                       // Toggle the 'E' pin to send the command.
   Delay1KTCYx(5);	
   PORTE &= ~E_PIN_MASK;
 
   TRISD = 0;
   PORTD = (cmdcode << 4);                     // Repeat for least significant nibble.
   PORTE  = LCD_CMD_WR;
   PORTE |= E_PIN_MASK;                       // Toggle the 'E' pin to send the command.
   Delay1KTCYx(5);
   PORTE &= ~E_PIN_MASK;

   lcd_wait();                                 // Call the wait routine.
 }
 
void wrdata (char data)
// Write a Character to the LCD Display. 
// In 4-bit mode we send the MSN first and then
// the LSN. We then call the wait routine to hold
// until the busy flag is cleared.
 { TRISD = 0;
   PORTD = data & 0xF0;                         // Get the most significant nibble first. 
   PORTE = LCD_DATA_WR;                         // Specify a data write operation.
   PORTE |= E_PIN_MASK;                         // Toggle the 'E' pin to send the command.
   Delay1KTCYx(5);
   PORTE &= ~E_PIN_MASK;
   
   TRISD = 0;
   PORTD = (data << 4);                         // Repeat for least significant nibble.
   PORTE = LCD_DATA_WR;
   PORTE |= E_PIN_MASK;
   Delay1KTCYx(5);
   PORTE &= ~E_PIN_MASK;
  
   lcd_wait();                                  // Call the wait routine.
 }

void wrcgchr(unsigned char *arrayptr, int offset)
// Subroutine to write a custom graphic character
// into CGRAM. We take a pointer to the array of bytes
// and an offset into CGRAM at which to place the character.
 { TRISD = PORTD = 0xFF;
   wrcmd(LCD_SETCGADDR + offset);               // Set the CG RAM address.
    
   while (*arrayptr != '\n') 
    { wrdata (*arrayptr++);
    }
 }

void lcd_wait ()
// Wait for the LCD busy flag to clear. 
// Crucially in 4-bit mode we must perform 2 read operations
// to retrieve both the high and the low nibble of the return
// byte. We then amalgamate the two nibbles and test the busy flag.
 { unsigned char lownibble = 0, highnibble = 0,status = 0;
   do
    { TRISD = 0xFF;                 
      PORTE = LCD_BUSY_RD;              
      PORTE |= E_PIN_MASK;
      Delay1KTCYx(5);
      highnibble = PORTD & 0xF0;                // read the high nibble.
      PORTE &= ~E_PIN_MASK;

      PORTE = TRISE = 0;                        // reset.
      TRISD = 0xFF; 
      PORTE = LCD_BUSY_RD;          
      PORTE |= E_PIN_MASK;
      Delay1KTCYx(5);
      lownibble = (PORTD & 0xF0) >> 4;          // read the low nibble. 
      PORTE &= ~E_PIN_MASK;
      status = highnibble + lownibble;          // combine to form busy status and ddaddress.
    } while (status & 0x80);                    // test busy flag.
 }

void lcd_write(char data)
{	// caien ia aaiie a aenieay
	wrdata (data);
}

void lcd_write_text(char text[], unsigned char br)
{	// caien ia oaeno a aenieay
	int i;
	for(i=0;i<br;i++)
	{
		if (!text[i])
			break;
		lcd_write(text[i]);
	Delay1KTCYx(50);
	}
}

void lcd_line(int line)
{	// iano?ieea ia ?aa a aenieay
	if (line==0)
	{
		wrcmd (0x80);
	}
	else 
	{
		wrcmd (0xC0);
	}
}